package coil.memory

import android.content.Context
import android.widget.ImageView
import androidx.test.core.app.ApplicationProvider
import coil.EventListener
import coil.ImageLoader
import coil.annotation.ExperimentalCoilApi
import coil.bitmappool.FakeBitmapPool
import coil.decode.DataSource
import coil.request.ErrorResult
import coil.request.RequestResult
import coil.request.SuccessResult
import coil.target.FakeTarget
import coil.target.ImageViewTarget
import coil.transition.Transition
import coil.transition.TransitionTarget
import coil.util.createBitmap
import coil.util.createGetRequest
import coil.util.createLoadRequest
import coil.util.createTestMainDispatcher
import coil.util.isInvalid
import coil.util.toDrawable
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.ExperimentalCoroutinesApi
import kotlinx.coroutines.delay
import kotlinx.coroutines.runBlocking
import kotlinx.coroutines.test.TestCoroutineDispatcher
import kotlinx.coroutines.test.resetMain
import org.junit.After
import org.junit.Before
import org.junit.Test
import org.junit.runner.RunWith
import org.robolectric.RobolectricTestRunner
import kotlin.test.assertFalse
import kotlin.test.assertTrue

@RunWith(RobolectricTestRunner::class)
@OptIn(ExperimentalCoilApi::class, ExperimentalCoroutinesApi::class)
class TargetDelegateTest {

    private lateinit var context: Context
    private lateinit var mainDispatcher: TestCoroutineDispatcher
    private lateinit var imageLoader: ImageLoader
    private lateinit var pool: FakeBitmapPool
    private lateinit var counter: BitmapReferenceCounter
    private lateinit var delegateService: DelegateService

    @Before
    fun before() {
        context = ApplicationProvider.getApplicationContext()
        mainDispatcher = createTestMainDispatcher()
        imageLoader = ImageLoader(context)
        pool = FakeBitmapPool()
        counter = BitmapReferenceCounter(EmptyWeakMemoryCache, pool, null)
        delegateService = DelegateService(imageLoader, counter, null)
    }

    @After
    fun after() {
        Dispatchers.resetMain()
        imageLoader.shutdown()
    }

    @Test
    fun `empty target does not invalidate`() {
        val request = createLoadRequest(context)
        val delegate = delegateService.createTargetDelegate(request, EventListener.NONE)

        runBlocking {
            val bitmap = createBitmap()
            val drawable = bitmap.toDrawable(context)
            delegate.start(drawable, drawable)
            assertFalse(counter.isInvalid(bitmap))
        }

        runBlocking {
            val bitmap = createBitmap()
            val result = SuccessResult(bitmap.toDrawable(context), DataSource.DISK)
            delegate.success(result, Transition.NONE)
            assertFalse(counter.isInvalid(bitmap))
        }
    }

    @Test
    fun `get request invalidates the success bitmap`() {
        val request = createGetRequest(context)
        val delegate = delegateService.createTargetDelegate(request, EventListener.NONE)

        runBlocking {
            val bitmap = createBitmap()
            val result = SuccessResult(bitmap.toDrawable(context), DataSource.DISK)
            delegate.success(result, Transition.NONE)
            assertTrue(counter.isInvalid(bitmap))
        }
    }

    @Test
    fun `target methods are called and bitmaps are invalidated`() {
        val target = FakeTarget()
        val request = createLoadRequest(context) {
            target(target)
        }
        val delegate = delegateService.createTargetDelegate(request, EventListener.NONE)

        runBlocking {
            val bitmap = createBitmap()
            val drawable = bitmap.toDrawable(context)
            delegate.start(drawable, null)
            assertTrue(target.start)
            assertTrue(counter.isInvalid(bitmap))
        }

        runBlocking {
            val bitmap = createBitmap()
            val result = SuccessResult(bitmap.toDrawable(context), DataSource.DISK)
            delegate.success(result, Transition.NONE)
            assertTrue(target.success)
            assertTrue(counter.isInvalid(bitmap))
        }

        runBlocking {
            val bitmap = createBitmap()
            val result = ErrorResult(bitmap.toDrawable(context), Throwable())
            delegate.error(result, Transition.NONE)
            assertTrue(target.error)
            assertFalse(counter.isInvalid(bitmap))
        }
    }

    @Test
    fun `request with poolable target returns previous bitmap to pool`() {
        val request = createLoadRequest(context) {
            target(ImageViewTarget(ImageView(context)))
        }
        val delegate = delegateService.createTargetDelegate(request, EventListener.NONE)

        val initialBitmap = createBitmap()
        val initialDrawable = initialBitmap.toDrawable(context)
        delegate.start(initialDrawable, initialDrawable)
        assertFalse(counter.isInvalid(initialBitmap))
        assertFalse(pool.bitmaps.contains(initialBitmap))

        runBlocking {
            val bitmap = createBitmap()
            val result = SuccessResult(bitmap.toDrawable(context), DataSource.DISK)
            delegate.success(result, Transition.NONE)
            assertFalse(counter.isInvalid(bitmap))
            assertTrue(pool.bitmaps.contains(initialBitmap))
        }
    }

    @Test
    fun `request suspends until transition is complete`() {
        val request = createLoadRequest(context) {
            target(ImageViewTarget(ImageView(context)))
        }
        val delegate = delegateService.createTargetDelegate(request, EventListener.NONE)

        val initialBitmap = createBitmap()
        val initialDrawable = initialBitmap.toDrawable(context)
        delegate.start(initialDrawable, initialDrawable)
        assertFalse(counter.isInvalid(initialBitmap))
        assertFalse(pool.bitmaps.contains(initialBitmap))

        runBlocking {
            val bitmap = createBitmap()
            var isRunning = true
            val result = SuccessResult(bitmap.toDrawable(context), DataSource.DISK)
            val transition = object : Transition {
                override suspend fun transition(target: TransitionTarget<*>, result: RequestResult) {
                    assertFalse(pool.bitmaps.contains(initialBitmap))
                    delay(100) // Simulate an animation.
                    assertFalse(pool.bitmaps.contains(initialBitmap))
                    isRunning = false
                }
            }
            delegate.success(result, transition)

            // Ensure that the animation completed and the initial bitmap was not pooled until this method completes.
            assertFalse(isRunning)
            assertTrue(pool.bitmaps.contains(initialBitmap))
        }
    }
}
